@@ -1,19 +1,23 @@
 """Access the Daymet database for both single single pixel and gridded queries."""
-from typing import Dict, Iterable, List, Optional, Tuple, Union
+import io
+from itertools import product
+from typing import Callable, Dict, Iterable, Iterator, List, Optional, Tuple, Union
 
 import numpy as np
 import pandas as pd
 import py3dep
 import pygeoogc as ogc
 import pygeoutils as geoutils
+import rasterio.transform as rio_transform
 import xarray as xr
 from pygeoogc import MatchCRS, RetrySession, ServiceURL
-from shapely.geometry import Polygon
+from shapely.geometry import MultiPolygon, Polygon
 
 from .exceptions import InvalidInputRange, InvalidInputType, InvalidInputValue, MissingItems
 
 DEF_CRS = "epsg:4326"
 DATE_FMT = "%Y-%m-%d"
+DATE_REQ = "%Y-%m-%dT%H:%M:%SZ"
 
 
 class Daymet:
@@ -30,10 +34,16 @@ class Daymet:
         Whether to compute evapotranspiration based on
         `UN-FAO 56 paper <http://www.fao.org/docrep/X0490E/X0490E00.htm>`__.
         The default is False
+    time_scale : str, optional
+        Data time scale which can be daily, monthly (monthly summaries),
+        or annual (annual summaries). Defaults to daily.
     """
 
     def __init__(
-        self, variables: Optional[Union[List[str], str]] = None, pet: bool = False,
+        self,
+        variables: Optional[Union[List[str], str]] = None,
+        pet: bool = False,
+        time_scale: str = "daily",
     ) -> None:
         self.session = RetrySession()
 
@@ -41,6 +51,13 @@ class Daymet:
 
         self.units = dict(zip(vars_table["Abbr"], vars_table["Units"]))
 
+        valid_times = ["daily", "monthly", "annual"]
+        if time_scale not in valid_times:
+            raise InvalidInputValue("time_scale", valid_times)
+
+        self.time_scale = time_scale
+        self.code = {"daily": 1840, "monthly": 1855, "annual": 1852}
+
         valid_variables = vars_table.Abbr.to_list()
         if variables is None:
             self.variables = valid_variables
@@ -54,6 +71,19 @@ class Daymet:
                 reqs = ("tmin", "tmax", "vp", "srad", "dayl")
                 self.variables = list(set(reqs) | set(self.variables))
 
+    @staticmethod
+    def check_dates(dates: Union[Tuple[str, str], Union[int, List[int]]]) -> None:
+        """Check if input dates are in correct format and valid."""
+        if not isinstance(dates, (tuple, list, int)):
+            raise InvalidInputType(
+                "dates", "tuple, list, or int", "(start, end), year, or [years, ...]"
+            )
+
+        if isinstance(dates, tuple) and len(dates) != 2:
+            raise InvalidInputType(
+                "dates", "Start and end should be passed as a tuple of length 2."
+            )
+
     @staticmethod
     def dates_todict(dates: Tuple[str, str]) -> Dict[str, str]:
         """Set dates by start and end dates as a tuple, (start, end)."""
@@ -119,7 +149,10 @@ class Daymet:
 
     @staticmethod
     def pet_byloc(
-        clm_df: pd.DataFrame, coords: Tuple[float, float], crs: str = DEF_CRS
+        clm_df: pd.DataFrame,
+        coords: Tuple[float, float],
+        crs: str = DEF_CRS,
+        alt_unit: bool = False,
     ) -> pd.DataFrame:
         """Compute Potential EvapoTranspiration using Daymet dataset for a single location.
 
@@ -137,30 +170,44 @@ class Daymet:
             Coordinates of the daymet data location as a tuple, (x, y).
         crs : str, optional
             The spatial reference of the input coordinate, defaults to epsg:4326
+        alt_unit : str, optional
+            Whether to use alternative units rather than the official ones, defaults to False.
 
         Returns
         -------
         pandas.DataFrame
             The input DataFrame with an additional column named ``pet (mm/day)``
         """
+        units = {
+            "dayl": ("s/day", "s"),
+            "srad": ("W/m2", "W/m^2"),
+            "tmax": ("degrees C", "deg c"),
+            "tmin": ("degrees C", "deg c"),
+        }
+
         va_pa = "vp (Pa)"
-        tmin_c = "tmin (deg c)"
-        tmax_c = "tmax (deg c)"
-        srad_wm2 = "srad (W/m^2)"
-        dayl_s = "dayl (s)"
+        tmin_c = f"tmin ({units['tmin'][alt_unit]})"
+        tmax_c = f"tmax ({units['tmax'][alt_unit]})"
+        srad_wm2 = f"srad ({units['srad'][alt_unit]})"
+        dayl_s = f"dayl ({units['dayl'][alt_unit]})"
         tmean_c = "tmean (deg c)"
 
         reqs = [tmin_c, tmax_c, va_pa, srad_wm2, dayl_s]
 
-        _check_requirements(reqs, clm_df)
+        _check_requirements(reqs, clm_df.columns)
 
         clm_df[tmean_c] = 0.5 * (clm_df[tmax_c] + clm_df[tmin_c])
         delta_v = (
             4098
-            * (0.6108 * np.exp(17.27 * clm_df[tmean_c] / (clm_df[tmean_c] + 237.3),))
+            * (
+                0.6108
+                * np.exp(
+                    17.27 * clm_df[tmean_c] / (clm_df[tmean_c] + 237.3),
+                )
+            )
             / ((clm_df[tmean_c] + 237.3) ** 2)
         )
-        elevation = py3dep.elevation_byloc(coords, crs)
+        elevation = py3dep.elevation_bycoords([coords], crs)[0]
 
         pa = 101.3 * ((293.0 - 0.0065 * elevation) / 293.0) ** 5.26
         gamma = pa * 0.665e-3
@@ -209,7 +256,7 @@ class Daymet:
         ) / (delta_v + gamma * (1 + 0.34 * u_2m))
         clm_df[va_pa] = clm_df[va_pa] * 1.0e3
 
-        return clm_df
+        return clm_df.drop(columns=tmean_c)
 
     @staticmethod
     def pet_bygrid(clm_ds: xr.Dataset) -> xr.Dataset:
@@ -239,16 +286,14 @@ class Daymet:
         dtype = clm_ds.tmin.dtype
         dates = clm_ds["time"]
         clm_ds["tmean"] = 0.5 * (clm_ds["tmax"] + clm_ds["tmin"])
-        clm_ds["tmean"].attrs["units"] = "degree C"
         clm_ds["delta_r"] = (
             4098
             * (0.6108 * np.exp(17.27 * clm_ds["tmean"] / (clm_ds["tmean"] + 237.3)))
             / ((clm_ds["tmean"] + 237.3) ** 2)
         )
 
-        gridxy = (clm_ds.x.values, clm_ds.y.values)
-        res = clm_ds.res[0] * 1000
-        elev = py3dep.elevation_bygrid(gridxy, clm_ds.crs, res)
+        res = clm_ds.res[0] * 1.0e3
+        elev = py3dep.elevation_bygrid(clm_ds.x.values, clm_ds.y.values, clm_ds.crs, res)
         attrs = clm_ds.attrs
         clm_ds = xr.merge([clm_ds, elev])
         clm_ds.attrs = attrs
@@ -307,7 +352,7 @@ class Daymet:
         clm_ds["time"] = dates
         clm_ds["vp"] *= 1.0e3
 
-        clm_ds = clm_ds.drop_vars(["delta_r", "gamma", "e_def", "rad_n"])
+        clm_ds = clm_ds.drop_vars(["delta_r", "gamma", "e_def", "rad_n", "tmean"])
 
         return clm_ds
 
@@ -321,6 +366,10 @@ def get_byloc(
 ) -> pd.DataFrame:
     """Get daily climate data from Daymet for a single point.
 
+    This function uses Daymet's RESTful service to get the daily
+    climate data and does not support monthly and annual summaries.
+    If you want to get the summaries directly use get_bycoords function.
+
     Parameters
     ----------
     coords : tuple
@@ -345,20 +394,19 @@ def get_byloc(
         Daily climate data for a location
     """
     daymet = Daymet(variables, pet)
+    daymet.check_dates(dates)
 
-    if isinstance(dates, tuple) and len(dates) == 2:
+    if isinstance(dates, tuple):
         dates_dict = daymet.dates_todict(dates)
-    elif isinstance(dates, (list, int)):
+    else:
         dates_dict = daymet.years_todict(dates)
-    else:
-        raise InvalidInputType("dates", "tuple or list", "(start, end) or [2001, 2010, ...]")
 
-    if isinstance(coords, tuple) and len(coords) == 2:
-        _coords = MatchCRS.coords(((coords[0],), (coords[1],)), crs, DEF_CRS)
-        lon, lat = (_coords[0][0], _coords[1][0])
-    else:
+    if not (isinstance(coords, tuple) and len(coords) == 2):
         raise InvalidInputType("coords", "tuple", "(lon, lat)")
 
+    _coords = MatchCRS.coords(((coords[0],), (coords[1],)), crs, DEF_CRS)
+    lon, lat = (_coords[0][0], _coords[1][0])
+
     if not ((14.5 < lat < 52.0) or (-131.0 < lon < -53.0)):
         raise InvalidInputRange(
             "The location is outside the Daymet dataset. "
@@ -381,25 +429,103 @@ def get_byloc(
     clm = clm.drop(["year", "yday"], axis=1)
 
     if pet:
-        clm = daymet.pet_byloc(clm, (lon, lat))
+        clm = daymet.pet_byloc(clm, (lon, lat), alt_unit=True)
+    return clm
+
+
+def get_bycoords(
+    coords: Tuple[float, float],
+    dates: Union[Tuple[str, str], Union[int, List[int]]],
+    loc_crs: str = DEF_CRS,
+    variables: Optional[List[str]] = None,
+    pet: bool = False,
+    region: str = "na",
+    time_scale: str = "daily",
+) -> xr.Dataset:
+    """Get point-data from the Daymet database at 1-km resolution.
+
+    This function uses THREDDS data service to get the coordinates
+    and supports getting monthly and annual summaries of the climate
+    data directly from the server.
+
+    Parameters
+    ----------
+    coords : tuple
+        Coordinates of the location of interest as a tuple (lon, lat)
+    dates : tuple or list, optional
+        Start and end dates as a tuple (start, end) or a list of years [2001, 2010, ...].
+    loc_crs : str, optional
+        The CRS of the input geometry, defaults to epsg:4326.
+    variables : str or list
+        List of variables to be downloaded. The acceptable variables are:
+        ``tmin``, ``tmax``, ``prcp``, ``srad``, ``vp``, ``swe``, ``dayl``
+        Descriptions can be found `here <https://daymet.ornl.gov/overview>`__.
+    pet : bool
+        Whether to compute evapotranspiration based on
+        `UN-FAO 56 paper <http://www.fao.org/docrep/X0490E/X0490E00.htm>`__.
+        The default is False
+    region : str, optional
+        Region in the US, defaults to na. Acceptable values are:
+        * na: Continental North America
+        * hi: Hawaii
+        * pr: Puerto Rico
+    time_scale : str, optional
+        Data time scale which can be daily, monthly (monthly summaries),
+        or annual (annual summaries). Defaults to daily.
+
+    Returns
+    -------
+    xarray.Dataset
+        Daily climate data within a geometry
+    """
+    daymet = Daymet(variables, pet, time_scale)
+    daymet.check_dates(dates)
+
+    if isinstance(dates, tuple):
+        dates_itr = daymet.dates_tolist(dates)
+    else:
+        dates_itr = daymet.years_tolist(dates)
+
+    _coords = MatchCRS.coords(((coords[0],), (coords[1],)), loc_crs, DEF_CRS)
+    coords = (_coords[0][0], _coords[1][0])
+    urls = coord_urls(daymet.code[time_scale], coords, region, daymet.variables, dates_itr)
+
+    clm = pd.concat(
+        (
+            pd.concat(
+                pd.read_csv(io.BytesIO(r), parse_dates=[0], usecols=[0, 3], index_col=[0])
+                for r in ogc.async_requests(u, "binary", max_workers=8)
+            )
+            for u in urls
+        ),
+        axis=1,
+    )
+    clm.columns = [c.replace('[unit="', " (").replace('"]', ")") for c in clm.columns]
+
+    if "prcp (mm)" in clm:
+        clm = clm.rename(columns={"prcp (mm)": "prcp (mm/day)"})
+
+    clm.index = pd.to_datetime(clm.index.strftime("%Y-%m-%d"))
+
+    if pet:
+        clm = daymet.pet_byloc(clm, coords, alt_unit=False)
     return clm
 
 
 def get_bygeom(
-    geometry: Union[Polygon, Tuple[float, float, float, float]],
+    geometry: Union[Polygon, MultiPolygon, Tuple[float, float, float, float]],
     dates: Union[Tuple[str, str], Union[int, List[int]]],
     geo_crs: str = DEF_CRS,
     variables: Optional[List[str]] = None,
     pet: bool = False,
-    fill_holes: bool = False,
+    region: str = "na",
+    time_scale: str = "daily",
 ) -> xr.Dataset:
-    """Gridded data from the Daymet database at 1-km resolution.
-
-    The data is clipped using NetCDF Subset Service.
+    """Get gridded data from the Daymet database at 1-km resolution.
 
     Parameters
     ----------
-    geometry : shapely.geometry.Polygon or bbox
+    geometry : Polygon, MultiPolygon, or bbox
         The geometry of the region of interest.
     dates : tuple or list, optional
         Start and end dates as a tuple (start, end) or a list of years [2001, 2010, ...].
@@ -413,63 +539,40 @@ def get_bygeom(
         Whether to compute evapotranspiration based on
         `UN-FAO 56 paper <http://www.fao.org/docrep/X0490E/X0490E00.htm>`__.
         The default is False
-    fill_holes : bool, optional
-        Whether to fill the holes in the geometry's interior, defaults to False.
+    region : str, optional
+        Region in the US, defaults to na. Acceptable values are:
+        * na: Continental North America
+        * hi: Hawaii
+        * pr: Puerto Rico
+    time_scale : str, optional
+        Data time scale which can be daily, monthly (monthly average),
+        or annual (annual average). Defaults to daily.
 
     Returns
     -------
     xarray.Dataset
         Daily climate data within a geometry
     """
-    daymet = Daymet(variables, pet)
+    daymet = Daymet(variables, pet, time_scale)
+    daymet.check_dates(dates)
 
-    if isinstance(dates, tuple) and len(dates) == 2:
+    if isinstance(dates, tuple):
         dates_itr = daymet.dates_tolist(dates)
-    elif isinstance(dates, (list, int)):
-        dates_itr = daymet.years_tolist(dates)
     else:
-        raise InvalidInputType("dates", "tuple or list", "(start, end) or [2001, 2010, ...]")
+        dates_itr = daymet.years_tolist(dates)
 
     _geometry = geoutils.geo2polygon(geometry, geo_crs, DEF_CRS)
-    _geometry = Polygon(_geometry.exterior) if fill_holes else _geometry
+    urls = gridded_urls(
+        daymet.code[time_scale], _geometry.bounds, region, daymet.variables, dates_itr
+    )
 
-    west, south, east, north = _geometry.bounds
-    base_url = ServiceURL().restful.daymet_grid
-    urls = []
-
-    for s, e in dates_itr:
-        for v in daymet.variables:
-            urls.append(
-                base_url
-                + "&".join(
-                    [
-                        f"{s.year}/daymet_v3_{v}_{s.year}_na.nc4?var=lat",
-                        "var=lon",
-                        f"var={v}",
-                        f"north={north}",
-                        f"west={west}",
-                        f"east={east}",
-                        f"south={south}",
-                        "disableProjSubset=on",
-                        "horizStride=1",
-                        f'time_start={s.strftime("%Y-%m-%dT%H:%M:%SZ")}',
-                        f'time_end={e.strftime("%Y-%m-%dT%H:%M:%SZ")}',
-                        "timeStride=1",
-                        "accept=netcdf",
-                    ]
-                )
-            )
-
-    def getter(url):
-        return xr.load_dataset(daymet.session.get(url).content)
-
-    data = xr.merge(ogc.utils.threading(getter, urls, max_workers=8))
+    clm = xr.open_mfdataset(ogc.async_requests(urls, "binary", max_workers=8))
 
     for k, v in daymet.units.items():
-        if k in data.variables:
-            data[k].attrs["units"] = v
+        if k in clm.variables:
+            clm[k].attrs["units"] = v
 
-    data = data.drop_vars(["lambert_conformal_conic"])
+    clm = clm.drop_vars(["lambert_conformal_conic"])
 
     crs = " ".join(
         [
@@ -485,33 +588,190 @@ def get_bygeom(
             "+no_defs",
         ]
     )
-    data.attrs["crs"] = crs
+    clm.attrs["crs"] = crs
+    clm.attrs["nodatavals"] = (0.0,)
 
-    x_res, y_res = data.x.diff("x").min().item(), data.y.diff("y").min().item()
-    # PixelAsArea Convention
-    x_origin = data.x.values[0] - x_res / 2.0
-    y_origin = data.y.values[0] - y_res / 2.0
+    xdim, ydim = "x", "y"
+    height, width = clm.sizes[ydim], clm.sizes[xdim]
 
-    transform = (x_res, 0, x_origin, 0, y_res, y_origin)
+    left, right = clm[xdim].min().item(), clm[xdim].max().item()
+    bottom, top = clm[ydim].min().item(), clm[ydim].max().item()
 
-    x_end = x_origin + data.dims["x"] * x_res
-    y_end = y_origin + data.dims["y"] * y_res
-    x_options = np.array([x_origin, x_end])
-    y_options = np.array([y_origin, y_end])
+    x_res = abs(left - right) / (width - 1)
+    y_res = abs(top - bottom) / (height - 1)
 
-    data.attrs["transform"] = transform
-    data.attrs["res"] = (x_res, y_res)
-    data.attrs["bounds"] = (
-        x_options.min(),
-        y_options.min(),
-        x_options.max(),
-        y_options.max(),
-    )
+    left -= x_res * 0.5
+    right += x_res * 0.5
+    top += y_res * 0.5
+    bottom -= y_res * 0.5
+
+    clm.attrs["transform"] = rio_transform.from_bounds(left, bottom, right, top, width, height)
+    clm.attrs["res"] = (x_res, y_res)
+    clm.attrs["bounds"] = (left, bottom, right, top)
 
     if pet:
-        data = daymet.pet_bygrid(data)
+        clm = daymet.pet_bygrid(clm)
 
-    return geoutils.xarray_geomask(data, _geometry, DEF_CRS)
+    if isinstance(clm, xr.Dataset):
+        for v in clm:
+            clm[v].attrs["crs"] = crs
+            clm[v].attrs["nodatavals"] = (0.0,)
+
+    return geoutils.xarray_geomask(clm, geometry, geo_crs)
+
+
+def get_filename(
+    code: int,
+    region: str,
+) -> Dict[int, Callable[[str], str]]:
+    """Generate an iterable URL list for downloading Daymet data.
+
+    Parameters
+    ----------
+    code : int
+        Endpoint code which should be one of the following:
+        * 1840: Daily
+        * 1855: Monthly average
+        * 1852: Annual average
+    region : str
+        Region in the US. Acceptable values are:
+        * na: Continental North America
+        * hi: Hawaii
+        * pr: Puerto Rico
+
+    Returns
+    -------
+    generator
+        An iterator of generated URLs.
+    """
+    valid_regions = ["na", "hi", "pr"]
+    if region not in valid_regions:
+        raise InvalidInputValue("region", valid_regions)
+
+    valid_codes = [1840, 1855, 1852]
+    if code not in valid_codes:
+        raise InvalidInputValue("code", valid_codes)
+
+    return {
+        1840: lambda v: f"daily_{region}_{v}",
+        1855: lambda v: f"{v}_monttl_{region}" if v == "prcp" else f"{v}_monavg_{region}",
+        1852: lambda v: f"{v}_annttl_{region}" if v == "prcp" else f"{v}_annavg_{region}",
+    }
+
+
+def coord_urls(
+    code: int,
+    coord: Tuple[float, float],
+    region: str,
+    variables: List[str],
+    dates: List[Tuple[pd.DatetimeIndex, pd.DatetimeIndex]],
+) -> Iterator[List[Tuple[str, Dict[str, str]]]]:
+    """Generate an iterable URL list for downloading Daymet data.
+
+    Parameters
+    ----------
+    code : int
+        Endpoint code which should be one of the following:
+        * 1840: Daily
+        * 1855: Monthly average
+        * 1852: Annual average
+    coord : tuple of length 2
+        Coordinates in EPSG:4326 CRS (lon, lat)
+    region : str
+        Region in the US. Acceptable values are:
+        * na: Continental North America
+        * hi: Hawaii
+        * pr: Puerto Rico
+    variables : list
+        A list of Daymet variables
+    date : list
+        A list of dates
+
+    Returns
+    -------
+    generator
+        An iterator of generated URLs.
+    """
+    time_scale = get_filename(code, region)
+
+    lon, lat = coord
+    base_url = f"{ServiceURL().restful.daymet}/{code}"
+    return (
+        [
+            (
+                f"{base_url}/daymet_v4_{time_scale[code](v)}_{s.year}.nc",
+                {
+                    "var": v,
+                    "longitude": f"{lon}",
+                    "latitude": f"{lat}",
+                    "time_start": s.strftime(DATE_REQ),
+                    "time_end": e.strftime(DATE_REQ),
+                    "accept": "csv",
+                },
+            )
+            for s, e in dates
+        ]
+        for v in variables
+    )
+
+
+def gridded_urls(
+    code: int,
+    bounds: Tuple[float, float, float, float],
+    region: str,
+    variables: List[str],
+    dates: List[Tuple[pd.DatetimeIndex, pd.DatetimeIndex]],
+) -> Iterator[Tuple[str, Dict[str, str]]]:
+    """Generate an iterable URL list for downloading Daymet data.
+
+    Parameters
+    ----------
+    code : int
+        Endpoint code which should be one of the following:
+        * 1840: Daily
+        * 1855: Monthly average
+        * 1852: Annual average
+    bounds : tuple of length 4
+        Bounding box (west, south, east, north)
+    region : str
+        Region in the US. Acceptable values are:
+        * na: Continental North America
+        * hi: Hawaii
+        * pr: Puerto Rico
+    variables : list
+        A list of Daymet variables
+    date : list
+        A list of dates
+
+    Returns
+    -------
+    generator
+        An iterator of generated URLs.
+    """
+    time_scale = get_filename(code, region)
+
+    west, south, east, north = bounds
+    base_url = f"{ServiceURL().restful.daymet}/{code}"
+    return (
+        (
+            f"{base_url}/daymet_v4_{time_scale[code](v)}_{s.year}.nc",
+            {
+                "var": v,
+                "north": f"{north}",
+                "west": f"{west}",
+                "east": f"{east}",
+                "south": f"{south}",
+                "disableProjSubset": "on",
+                "horizStride": "1",
+                "time_start": s.strftime(DATE_REQ),
+                "time_end": e.strftime(DATE_REQ),
+                "timeStride": "1",
+                "addLatLon": "true",
+                "accept": "netcdf",
+            },
+        )
+        for v, (s, e) in product(variables, dates)
+    )
 
 
 def _check_requirements(reqs: Iterable, cols: List[str]) -> None:
